#!/usr/bin/env python3

# Extract trace gas profile(s) from a CTMANA or CTMFCT file. 
# 
# There are two modes for this tool: 
#
# 1. extract for a given time and geolocation (single point)
# 2. extract all profiles for a level 2 granule and add it to the L2 data file;
#    this requires that you have write access to the L2 data file. 
#
# One of the species can be selected: NO2, SO2 or HCHO.

__version__ = "1.0.1"
__date__ = "February 2022" 
DownloadURL = "http://www.tropomi.eu/data-products/nitrogen-dioxide/"
 
# =====================================        

import os
import sys
import argparse
import datetime
from requests.structures import CaseInsensitiveDict

import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

import inspect

import numpy as np
import netCDF4
from scipy import constants

try:
    # using numba will speed up the code significantly, but it is a non-standard package.
    from numba import jit

    @jit(nopython=True)
    def calc_field(weights, spatial_index_lat, spatial_index_lon,
                   surface_pressure_external, surface_pressure, field,
                   result_field, result_pressure, hya, hyb):
        for pix_idx in range(weights.shape[1]):
            if surface_pressure_external is not None:
                ps = surface_pressure_external.flat[pix_idx]
            else:
                ps = 0.0 
            for i in range(4):
                if surface_pressure_external is None:
                    ps += (weights[i, pix_idx] * 
                            surface_pressure[spatial_index_lat[i, pix_idx], 
                                             spatial_index_lon[i, pix_idx]])
                result_field[:, pix_idx] += (weights[i, pix_idx] * 
                                             field[:, spatial_index_lat[i, pix_idx], 
                                             spatial_index_lon[i, pix_idx]])

            result_pressure[:, pix_idx] = hya + ps * hyb

        return result_field, result_pressure
except ImportError:
    print("The numba module is not available, running without.\nInstalling numba may increase the speed of this code by a factor 180.", file=sys.stderr)
    def calc_field(weights, spatial_index_lat, spatial_index_lon,
                   surface_pressure_external, surface_pressure, field,
                   result_field, result_pressure, hya, hyb):
        for pix_idx in range(weights.shape[1]):
            if surface_pressure_external is not None:
                ps = surface_pressure_external.flat[pix_idx]
            else:
                ps = 0.0 
            for i in range(4):
                if surface_pressure_external is None:
                    ps += (weights[i, pix_idx] * 
                            surface_pressure[spatial_index_lat[i, pix_idx], 
                                             spatial_index_lon[i, pix_idx]])
                result_field[:, pix_idx] += (weights[i, pix_idx] * 
                                             field[:, spatial_index_lat[i, pix_idx], 
                                             spatial_index_lon[i, pix_idx]])

            result_pressure[:, pix_idx] = hya + ps * hyb

        return result_field, result_pressure


named_geolocation_points = CaseInsensitiveDict({
    'KNMI':         {'lon': 5.177999,   'lat': 52.101544},
    'Cabauw':       {'lon': 4.926318,   'lat': 51.970268},
    'BIRA':         {'lon': 50.796781,  'lat': 4.356972},
    'IUP':          {'lon': 8.84964,    'lat': 53.103718},
    'DLR':          {'lon': 11.276202,  'lat': 48.084865},
    'ESTEC':        {'lon': 4.42015,    'lat': 52.217917},
    'ESRIN':        {'lon': 12.671675,  'lat': 41.827141},
    'RAL':          {'lon': -1.310875,  'lat': 51.576541},
    'FMI':          {'lon': 24.961061,  'lat': 60.203779},
    'MPI':          {'lon': 8.228973,   'lat': 49.991104},
    'Sodankyla':    {'lon': 26.588973,  'lat': 67.415999},
    'GSFC':         {'lon': -76.839854, 'lat': 38.992783},
    'Paramaribo':   {'lon': -55.204717, 'lat': 5.449259},
    'Jungfraujoch': {'lon': 7.985249,   'lat': 46.547463},
    'Xianghe':      {'lon': 116.960,    'lat': 39.750}})

# =====================================        

class CTM:
    def __init__(self, fname:str, species:str='no2'):
        self.fname = fname
        self.slice_index = None
        self.weights = None
        self.spatial_index_lon = None
        self.spatial_index_lat = None
        self.species = species
        self.final_output_shape = None
        self.ref = None
    
    def __enter__(self): 
        self.ref = netCDF4.Dataset(self.fname, 'r')
        return self.ref
  
    def __exit__(self, exception_type, exception_value, traceback): 
        if self.ref is not None:
            self.ref.close()
            self.ref = None
    
    def close(self):
        self.__exit__(None, None, None)
    
    def find_slice(self, t: datetime.datetime) -> int:
        """Find the correct time index"""
        time_ref = self.ref["time"]
        times = time_ref[:]
        time_units = time_ref.units
        
        try:
            f_times = netCDF4.num2date(times, time_units, 
                        only_use_cftime_datetimes=False, 
                        only_use_python_datetimes=True)
        except TypeError:
            f_times = netCDF4.num2date(times, time_units)
        
        quarter = datetime.timedelta(minutes=15)
        second  = datetime.timedelta(seconds=1)
        extra = quarter - second

        if t < f_times[0] - extra or t > f_times[-1] + extra:
            raise RuntimeError("Requested time not covered in CTM file.")
        else:
            self.slice_index = np.argmin([np.fabs((t - tt).total_seconds()) for tt in f_times])
        self.slice_time = f_times[self.slice_index]
        return self.slice_index,self.slice_time
        
    def find_indices_weights(self, latitude, longitude):
        """Find indices and weights for the spatial interpolation

Result applies to flattened array
"""
        if isinstance(latitude, list):
            latitude = np.asarray(latitude, dtype=np.float64)
            longitude = np.asarray(longitude, dtype=np.float64)
        elif isinstance(latitude, (int, float)):
            latitude = np.asarray([latitude], dtype=np.float64)
            longitude = np.asarray([longitude], dtype=np.float64)
        
        if np.any(latitude > 90.0) or np.any(latitude < -90.0):
            raise RuntimeError("Latitude out of bounds")
        if np.prod(latitude.shape) != np.prod(longitude.shape):
            raise RuntimeError("Size of latitude and longitude must match")
        
        self.final_output_shape = list(self.ref['hyam'].shape)
        self.final_output_shape.extend(list(latitude.shape))
        
        latitude = latitude.flatten()
        longitude = longitude.flatten()
        
        # size of TM5 array
        jm, im = self.ref[self.species].shape[-2:]
        
        longitude = np.where(longitude < -180.0, longitude + 360.0, longitude)
        longitude = np.where(longitude >= 180.0, longitude - 360.0, longitude)
        dlon = 360.0/im
        firstcellmid = 0.5*dlon-180.0
        rx = (longitude-firstcellmid)/dlon
        ix = np.asarray(rx, dtype=np.int32)
        gdx = rx - ix
        ix = np.where(ix == im, 0, ix)
        
        dlat = 180.0/jm
        firstcellmid = 0.5*dlat-90.0
        ry = (latitude-firstcellmid)/dlat
        iy = np.asarray(ry, dtype=np.int32)
        gdy = ry - iy
        
        # the other corner
        ix2 = ix + 1
        ix2 = np.where(ix2 == im, 0, ix2)
        iy2 = iy + 1
        if np.any(iy2 >= jm):
            idy = iy2 >= jm
            iy[idy] = jm-2
            iy2[idy] = jm-1
            gdy[idy] = 1.0
        if np.any(iy < 0):
            idy = iy < 0
            iy[idy] = 0
            iy2[idy] = 1
            gdy[idy] = 1.0
        
        output_shape = [4]
        output_shape.extend(list(latitude.shape))
        output_shape = tuple(output_shape)
        ix4a = np.concatenate((ix, ix2, ix, ix2)).reshape(output_shape)
        iy4a = np.concatenate((iy, iy, iy2, iy2)).reshape(output_shape)
        w4a = np.concatenate(((1-gdx)*(1-gdy), gdx*(1-gdy), (1-gdx)*gdy, gdx*gdy)).reshape(output_shape)
        
        self.weights = w4a
        self.spatial_index_lon = ix4a
        self.spatial_index_lat = iy4a

        return ix4a, iy4a, w4a
            
    def get_pressures_interfaces(self):
        """Get pressure at the interfaces of the model"""
        field = self.ref[self.species]
        pressure_shape = list(field.shape[1:])
        pressure_shape[0] += 1
        p_interfaces = np.zeros(pressure_shape, dtype=np.float32)
        ps = self.ref['ps'][self.slice_index, ...]
        hyai = self.ref['hyai'][:]
        hybi = self.ref['hybi'][:]
        
        for i, (a, b) in enumerate(zip(hyai, hybi)):
            p_interfaces[i, ...] = a + b*ps
            
        return p_interfaces

    def get_partial_column_field(self, p_interfaces=None):
        """Translate input field in volume mixing ratio to mol/m?? per layer"""
        xmair = 28.94 # mass of air (gram/mol)
        # conversion factor from mixing ratio, pressure drop (in Pa) to columns (in 10^15 molec cm^-2)
        pdiff2moleccm2 = 1.0e-15 * 0.1 * constants.Avogadro/(constants.g * xmair)
        # conversion factor from mixing ratio, pressure drop (in Pa) to columns in mol/m2.
        pdiffmolm2 =  1000 /(constants.g * xmair)
        
        field = self.ref[self.species]
        ps = self.ref['ps']
        hyai = self.ref['hyai'][:]
        hybi = self.ref['hybi'][:]
        
        # partial columns
        partial_column_field = np.zeros(field.shape[1:], dtype=np.float32)
        if p_interfaces is None:
            p_interfaces = self.get_pressures_interfaces()
        
        for level in range(self.ref[self.species].shape[1]):
            dp = p_interfaces[level, ...] - p_interfaces[level+1, ...]
            partial_column_field[level, ...] = pdiffmolm2 * dp * field[self.slice_index, level, ...]
        
        return partial_column_field
    
    def get_tropopause_pressure(self, surface_pressure_external=None):
        t_idx = self.slice_index
        hyai = self.ref['hyai'][:]
        hybi = self.ref['hybi'][:]
        surface_pressure = self.ref["ps"][t_idx, ...]
        tropopause_layer_index = self.ref["tropopause_layer_index"][t_idx, ...]
        p_interfaces = self.get_pressures_interfaces()
        dest_shape = (self.weights.shape[1], )
        p_tropo = np.zeros(dest_shape, dtype=np.float32)
        
        for pix_idx in range(self.weights.shape[1]):
            if surface_pressure_external is not None:
                ps = surface_pressure_external.flat[pix_idx]
            else:
                ps = 0.0
                for i in range(4):
                    ps += (self.weights[i, pix_idx] * 
                            surface_pressure[self.spatial_index_lat[i, pix_idx], 
                                             self.spatial_index_lon[i, pix_idx]])
                    
            for i in range(4):
                idx = tropopause_layer_index[self.spatial_index_lat[i, pix_idx], 
                                             self.spatial_index_lon[i, pix_idx]] + 1 # top of layer
                
                p_tropo[pix_idx] += self.weights[i, pix_idx] * (hyai[idx] + ps * hybi[idx])
        
        p_tropo = p_tropo.reshape(self.final_output_shape[1:])
        return p_tropo
        
    def get_temperature(self, surface_pressure_external=None):
        t_idx = self.slice_index
        p_interfaces = self.get_pressures_interfaces()
        field = self.ref["t"][t_idx, ...]
        surface_pressure = self.ref["ps"][t_idx, ...]
        hyam = self.ref['hyam'][:]
        hybm = self.ref['hybm'][:]
        
        dest_shape = (field.shape[0], self.weights.shape[1])
        result_field = np.zeros(dest_shape, dtype=np.float32)
        result_pressure = np.zeros(dest_shape, dtype=np.float32)
        
        result_field, result_pressure = calc_field(self.weights, self.spatial_index_lat, self.spatial_index_lon,
                                                   surface_pressure_external, surface_pressure, field,
                                                   result_field, result_pressure, hyam, hybm)
        return result_field, result_pressure
        
    def get_profiles(self, vmr=True, surface_pressure_external=None):
        t_idx = self.slice_index
        
        p_interfaces = self.get_pressures_interfaces()
        if not vmr:
            field = self.get_partial_column_field(p_interfaces)
        else:
            field = self.ref[self.species][t_idx, ...]
        surface_pressure = self.ref["ps"][t_idx, ...]
        hyai = self.ref['hyai'][:]
        hybi = self.ref['hybi'][:]
        
        dest_shape = (field.shape[0], self.weights.shape[1])
        result_field = np.zeros(dest_shape, dtype=np.float32)
        pressure_shape = (field.shape[0]+1, self.weights.shape[1])
        result_pressure = np.zeros(pressure_shape, dtype=np.float32)

        result_field, result_pressure = calc_field(self.weights, self.spatial_index_lat, self.spatial_index_lon,
                                                   surface_pressure_external, surface_pressure, field,
                                                   result_field, result_pressure, hyai, hybi)
                
        result_field = result_field.reshape(self.final_output_shape)
        pressure_shape = list(self.final_output_shape)
        pressure_shape[0] += 1
        result_pressure = result_pressure.reshape(tuple(pressure_shape))
        
        return result_field, result_pressure
        
def get_function_name():
   return inspect.stack()[1][3]


def extract_single(species:str='no2', ctm:str=None, vmr:bool=False, 
        time:datetime.datetime=None, geo_point:tuple=None, 
        pressure:float=None, verbose:int=0, **kwargs):
    ctmobj = CTM(ctm, species)
    with ctmobj:
        try:
            ctmobj.find_slice(time)
        except Exception as err:
            print(err, file=sys.stderr)
            sys.exit(1)
        ctmobj.find_indices_weights(geo_point[1], geo_point[0])
        if pressure is None:
            pres_array = None
        else:
            pres_array = np.asarray([100*pressure]) # hPa -> Pa.
        result_field, result_pressure  = ctmobj.get_profiles(vmr, surface_pressure_external=pres_array)
        result_temperature, result_pressure_mid = ctmobj.get_temperature(pres_array)
        
        tropopause_pressure = ctmobj.get_tropopause_pressure(pres_array)
    
    if vmr:
        print(f"pressure_bottom\tpressure_top\tpressure_mid\ttemperature\t{species}_vmr\n# hPa\t# hPa \t# hPa\t# K\t# 1")
    else:
        print(f"pressure_bottom\tpressure_top\tpressure_mid\ttemperature\t{species}_partial_columns\n# hPa\t# hPa \t# hPa\t# K\t# mol/m2 in layer")
    print(f"# time_slice: {ctmobj.slice_time:%Y-%m-%dT%H:%M:%SZ}, latitude: {geo_point[1]}, longitude: {geo_point[0]}")
    print(f"# requested_time: {time:%Y-%m-%dT%H:%M:%SZ}")
    print(f"# source: {os.path.basename(ctm)}")
    print(f"# tropopause pressure: {tropopause_pressure[0]/100} hPa")
    
    for pres_bot, pres_top, pres_mid, t, val in zip(result_pressure[0:-1], result_pressure[1:], result_pressure_mid, result_temperature, result_field):
        print(f"{pres_bot[0]/100}\t{pres_top[0]/100}\t{pres_mid[0]/100}\t{t[0]}\t{val[0]}")
    
    
def extract_orbit(species:str='no2', ctm:str=None, vmr:bool=False,
        l2_file:str=None, use_TM5_pressure:bool=False, verbose:int=0, 
        add_temperature_profile:bool=False, **kwargs):
    
    if not os.access(l2_file, os.O_RDWR):
        print(f"""Error when opening '{os.path.basename(l2_file)}' in append mode. 
We need write permission to add the model profiles to the level 2 file.""", file=sys.stderr)
        return
    
    ctmobj = CTM(ctm, species)
    ref = netCDF4.Dataset(l2_file, 'a')
    with ref, ctmobj:
        tstart = datetime.datetime.strptime(ref.time_coverage_start[0:19], '%Y-%m-%dT%H:%M:%S')
        tend = datetime.datetime.strptime(ref.time_coverage_end[0:19], '%Y-%m-%dT%H:%M:%S')
        time = tstart + (tend-tstart)/2
        
        if verbose > 0:
            print("Reading L2 data", file=sys.stderr)
            
        latitudes = ref['/PRODUCT/latitude'][:]
        longitudes = ref['/PRODUCT/longitude'][:]
        if use_TM5_pressure:
            pressures = None
        else:
            pressures = ref['/PRODUCT/SUPPORT_DATA/INPUT_DATA/surface_pressure'][:]
            
        try:
            ctmobj.find_slice(time)
        except Exception as err:
            print(err, file=sys.stderr)
            sys.exit(1)
        
        if verbose > 0:
            print("Calculating weights", file=sys.stderr)
        ctmobj.find_indices_weights(latitudes, longitudes)
        
        if verbose > 0:
            print("Retrieving profiles", file=sys.stderr)
        result_field, result_pressure = ctmobj.get_profiles(vmr,
            surface_pressure_external=pressures)
        
        if verbose > 1:
            print(f"""NO2 dimensions: {result_field.shape}
pressure bounds dimensions: {result_pressure.shape}""")
            print("Writing output", file=sys.stderr)

        name_map = {'no2': 'nitrogen_dioxide', 'so2': 'sulfur_dioxide', 'hcho': 'formaldehyde'}
        grp = ref['/PRODUCT/SUPPORT_DATA/DETAILED_RESULTS']
        vname = f"{species}_vmr" if vmr else f"{species}_partial_columns"
        try:
            var = grp.createVariable(vname, np.float32,
                                    dimensions=('time', 'scanline', 'ground_pixel', 'layer'),
                                    zlib=True, fletcher32=True)
        except RuntimeError:
            var = grp[vname]
            
        var[:] = result_field.transpose((1,2,3,0))
        var.units = "1" if vmr else "mol/m2"
        var.long_name = (f"TM5-MP a-priori profile of mixing ratios of {species.upper()} co-located with the observations" 
                         if vmr else 
                         f"TM5-MP a-priori profile of model grid-cell partial columns of {species.upper()} co-located with the observations")
        var.standard_name = (f"mole_fraction_of_{name_map[species]}_in_air" 
                             if vmr else 
                             f"mole_content_of_{name_map[species]}_in_atmosphere_layer")
        var.coordinates = "/PRODUCT/longitude /PRODUCT/latitude"
        if use_TM5_pressure:
            var.ancillary_variables = "/PRODUCT/tm5_constant_a /PRODUCT/tm5_constant_b /PRODUCT/tm5_tropopause_layer_index /PRODUCT/SUPPORT_DATA/INPUT_DATA/TM5_surface_pressure"
        else:
            var.ancillary_variables = "/PRODUCT/tm5_constant_a /PRODUCT/tm5_constant_b /PRODUCT/tm5_tropopause_layer_index /PRODUCT/SUPPORT_DATA/INPUT_DATA/surface_pressure"
        
        if use_TM5_pressure:
            try:
                var = grp.createVariable("TM5_surface_pressure", np.float32,
                                    dimensions=('time', 'scanline', 'ground_pixel'),
                                    zlib=True, fletcher32=True)
            except RuntimeError:
                var = grp["TM5_surface_pressure"]
            var[:] = result_pressure[0, ...]
            var.units = "Pa"
            var.long_name = f"Surface pressure used to calculate pressure grid for {vname} profiles"
            var.standard_name = 'surface_air_pressure'
            var.coordinates = "/PRODUCT/longitude /PRODUCT/latitude"
        
        if add_temperature_profile:
            if verbose > 0:
                print("Extracting temperatures", file=sys.stderr)
            result_temperature, result_pressure_mid = ctmobj.get_temperature()
            if verbose > 1:
                print(f"""temperature dimensions: {result_temperature.shape}
mid pressure dimensions: {result_pressure_mid.shape}""", file=sys.stderr)
            if verbose > 0:
                print("Writing temperatures", file=sys.stderr)
            vname = "temperature"
            try:
                var = grp.createVariable(vname, np.float32,
                                    dimensions=('time', 'scanline', 'ground_pixel', 'layer'),
                                    zlib=True, fletcher32=True)
            except RuntimeError:
                var = grp[vname]
            var[:] = result_temperature.transpose((1,2,3,0))
            var.units = "K"
            var.long_name = "TM5-MP temperature profile"
            var.standard_name = "air_temperature"
            var.coordinates = "/PRODUCT/longitude /PRODUCT/latitude"
            if use_TM5_pressure:
                var.ancillary_variables = "/PRODUCT/tm5_constant_a /PRODUCT/tm5_constant_b /PRODUCT/tm5_tropopause_layer_index /PRODUCT/SUPPORT_DATA/INPUT_DATA/TM5_surface_pressure"
            else:
                var.ancillary_variables = "/PRODUCT/tm5_constant_a /PRODUCT/tm5_constant_b /PRODUCT/tm5_tropopause_layer_index /PRODUCT/SUPPORT_DATA/INPUT_DATA/surface_pressure"
    
def argparse_datetime(s: str):
    try:
        value = datetime.datetime.strptime(s, "%Y-%m-%dT%H:%M:%S")
    except ValueError as err:
        msg = "{0} does not match format '%Y-%m-%dT%H:%M:%S'".format(s)
        raise argparse.ArgumentTypeError(msg)
    return value

def geospec_verify(s: str):
    global named_geolocation_points
    if s not in named_geolocation_points:
        names = "', '".join(sorted(list(named_geolocation_points.keys())))
        msg = "'{0}' not found. \nValid point names are '{1}'.".format(s, names)
        raise argparse.ArgumentTypeError(msg)
    return s

def geo_point_name(name: str):
    global named_geolocation_points
    # https://mynasadata.larc.nasa.gov/latitudelongitude-finder/
    if name in named_geolocation_points:
        return (named_geolocation_points[name]['lon'], named_geolocation_points[name]['lat'])
    else:
        names = "', '".join(sorted(list(named_geolocation_points.keys())))
        raise RuntimeError("Warning, '{0}' not found. \nValid point names are '{1}'.".format(name, names))

# =====================================        

def main():
    global named_geolocation_points
    geolocation_point_name_list = list(named_geolocation_points.keys())
    geolocation_point_name_list.sort()
    geolocation_point_names = "', '".join(geolocation_point_name_list)
    
    parser = argparse.ArgumentParser(description='Get CTM profile(s) for NO2, SO2 or HCHO.',
                                     epilog="This program can be downloaded from {0}".format(DownloadURL))
    
    # Main parser arguments
    
    parser.add_argument('--version', action='version', version='%(prog)s {0} ({1})'.format(__version__, __date__), 
                        help='show version info and exit')
    
    parser.add_argument('-v', '--verbose', action='count', default=0, help="be chatty")
    
    parser.add_argument('-s', '--species', metavar='MOLECULE', 
                        choices=('no2', 'so2', 'ch2o', 'hcho', 'NO2', 'SO2', 'CH2O', 'HCHO', 'NO???', 'SO???', 'CH???O'),
                        default='no2', 
                        help='trace gas for which to extract the profil, default: NO2; valid arguments: NO2, SO2, HCHO, CH2O')
    
    parser.add_argument('-c', '--ctm', required=True, help='the CTM file with the profiles')
    
    parser.add_argument('-V', '--vmr', action='store_true',
                        help="store the profile as vmr rather than partial columns")
    
    subparsers = parser.add_subparsers(help='Processing mode (these have their own help)', required=True)
    
    # Subparser arguments for "single" mode
     
    parser_s = subparsers.add_parser('single', help='Extract the profile for a single time and location.')
    
    parser_s.set_defaults(func=extract_single)
    parser_s.add_argument('-t', '--time', metavar='YYYY-MM-DDTHH:MM:SS', type=argparse_datetime,
                          required=True,
                          help='date and time to extract the profile for')
    group = parser_s.add_mutually_exclusive_group(required=True)
    group.add_argument('-g', '--geo-point', metavar="L", type=float, dest="geo_point", 
                          nargs=2, 
                          help="location to extract the profile for as a LAT LON pair")
    group.add_argument('-G', '--geo-name', metavar="NAME", type=geospec_verify, dest="geo_point_name", 
                          help="location to extract a profile for by name; recognized names are: '{0}'.".format(geolocation_point_names))
    parser_s.add_argument('-p', '--pressure', type=float, default=None, 
                          help="surface pressure in hPa to use for the profile; if not given the TM5 surface pressure is used")
    
    # Subparser arguments for "level2" mode
    
    parser_m = subparsers.add_parser('level2', help='Extract profiles for a full level-2 file; the profiles will be added to the file (for this mode read/write access to the level-2 file is required).')
    
    parser_m.set_defaults(func=extract_orbit)
    parser_m.add_argument('-i', '--input', metavar='L2_FILE', dest='l2_file', required=True, help='level-2 file to add the profile data to')
    parser_m.add_argument('--use-TM5-pressure', action='store_true',
                          help="use the surface pressure from TM5 rather than the surface pressure given in the level-2 file")
    parser_m.add_argument('--add-temperature-profile', action='store_true',
                          help="also add the temperature profile from TM5 to the level-2 file")
    
    # Check the parser argements
    
    args = vars(parser.parse_args())
    
    processor = args["func"]
    del args["func"]
        
    if "geo_point_name" in args and args["geo_point_name"]:
        args["geo_point"] = geo_point_name(args["geo_point_name"])
        del args["geo_point_name"]
    elif "geo_point" in args and args["geo_point"]:
        args["geo_point"] = args["geo_point"][1], args["geo_point"][0]
    
    if args["species"] in ("NO2", "no2", "NO???"):
        args["species"] = "no2"
    elif args["species"] in ("SO2", "so2", "SO???"):
        args["species"] = "so2"
    else:
        args["species"] = "ch2o"
    
    # Call the processor
    
    processor(**args)
    
    
if __name__ == "__main__":
    main()

